# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import logging

LOG = logging.getLogger(__name__)


class Resolver:

    def __init__(self, cloud, downloader, output_path):
        self.cloud = cloud
        self._memo = set()
        self._downloader = downloader
        self._output_path = output_path
        self._uuid_map_file = 'uuids.csv'
        self._var_counter = itertools.count(1)

    def _mk_var_name(self, slug):
        name = '{}{:03d}'.format(slug, next(self._var_counter))
        self._last_var = name
        return name

    def _map_uuids(self, resource_type, name, old_id, var_expr):
        pat = '"{}","{}","{}"'.format(
            resource_type, name, old_id,
        )
        line = '{},"{}"'.format(
            pat, '{{' + self._last_var + '.' + var_expr + '}}',
        )
        return {
            'name': 'Map UUID for {} {}'.format(resource_type, name),
            'lineinfile': {
                'dest': self._uuid_map_file,
                'state': 'present',
                'regexp': pat,
                'insertafter': 'EOF',
                'line': line,
            },
        }

    def init_tasks(self):
        yield {
            'name': 'Initializing UUID mapping file',
            'lineinfile': {
                'create': 'yes',
                'dest': self._uuid_map_file,
                'state': 'present',
                'regexp': '"Resource Type","Resource Name","Old","New"',
                'insertbefore': 'BOF',
                'line': '"Resource Type","Resource Name","Old","New"',
            },
        }

    def keypair(self, keypair):
        if ('keypair', keypair.name) in self._memo:
            return
        self._memo.add(('keypair', keypair.name))
        yield {
            'name': 'Add public key {}'.format(keypair.name),
            'os_keypair': {
                'name': keypair.name,
                'public_key': keypair.public_key,
            },
            'register': self._mk_var_name('key'),
        }
        yield self._map_uuids(
            'keypair', keypair.name, keypair.id, 'key.id',
        )

    def security_group(self, group):
        if ('security_group', group.id) in self._memo:
            return
        self._memo.add(('security_group', group.id))
        remote_groups = {}
        yield {
            'name': 'Add security group {}'.format(group.name),
            'os_security_group': {
                'state': 'present',
                'name': group.name,
                'description': group.description,
            },
            'register': self._mk_var_name('sg'),
        }
        yield self._map_uuids(
            'security group', group.name, group.id,
            'secgroup.id',
        )
        for rule in group.security_group_rules:
            rule_data = {
                'direction': rule.direction,
                'ethertype': rule.ethertype,
                'protocol': rule.protocol,
                'security_group': group.name,
                'state': 'present',
            }
            for opt in ['port_range_min', 'port_range_max',
                        'remote_group_id', 'remote_ip_prefix']:
                value = rule.get(opt)
                if value:
                    rule_data[opt] = value
            if rule.remote_group_id:
                if rule.remote_group_id not in remote_groups:
                    remote_group = self.cloud.get_security_group(rule.remote_group_id)
                    remote_groups[remote_group.id] = remote_group.name
                rule_data['remote_group_id'] = remote_groups[rule.remote_group_id]
            yield {
                'name': 'Add rule to security group {}'.format(group.name),
                'os_security_group_rule': rule_data,
            }

    def volume(self, volume, save_state):
        if ('volume', volume.name) in self._memo:
            return
        self._memo.add(('volume', volume.name))
        if save_state:
            image_name = 'volume-capture-{}'.format(volume.id)
            image_info = self.cloud.get_image(image_name)
            if image_info:
                LOG.info('found existing capture of volume %s in %s',
                         volume.name, image_name)
            else:
                LOG.info('creating image from volume %s as %s',
                         volume.name, image_name)
                image_info = self.cloud.create_image(
                    name=image_name,
                    volume=volume.id,
                    wait=True,
                )
            yield from self.image(image_info)
        os_volume = {
            'display_name': volume.display_name,
            'display_description': volume.display_description,
            'size': volume.size,
            'state': 'present',
        }
        if save_state:
            os_volume['image'] = image_name
        yield {
            'name': 'Create volume {}'.format(volume.name),
            'os_volume': os_volume,
            'register': self._mk_var_name('vol'),
        }
        yield self._map_uuids('volume', volume.name, volume.id, 'volume.id')

    def network(self, network):
        if ('network', network.name) in self._memo:
            return
        self._memo.add(('network', network.name))
        yield {
            'name': 'Create network {}'.format(network.name),
            'os_network': {
                'name': network.name,
                'external': network.get('router:external', False),
                'shared': network.shared,
                'state': 'present',
            },
            'register': self._mk_var_name('net'),
        }
        yield self._map_uuids('network', network.name, network.id, 'network.id')
        for subnet_id in network.subnets:
            subnet = self.cloud.get_subnet(subnet_id)
            os_subnet = {
                'name': subnet.name,
                'network_name': network.name,
                'cidr': subnet.cidr,
                'state': 'present',
                'allocation_pool_start': subnet.allocation_pools[0]['start'],
                'allocation_pool_end': subnet.allocation_pools[0]['end'],
                'dns_nameservers': subnet.dns_nameservers,
                'ip_version': subnet.ip_version,
            }
            if subnet.gateway_ip:
                os_subnet['gateway_ip'] = subnet.gateway_ip
            if subnet.ip_version == 6:
                os_subnet['ipv6_ra_mode'] = subnet.ipv6_ra_mode
                os_subnet['ipv6_address_mode'] = subnet.ipv6_address_mode
            yield {
                'name': 'Create subnet {} on network {}'.format(subnet.name,
                                                                network.name),
                'os_subnet': os_subnet,
                'register': self._mk_var_name('subnet'),
            }
            yield self._map_uuids('subnet', subnet.name, subnet.id, 'subnet.id')

    def server(self, server, save_state, key_name=None):
        # The ssh key needed to login to the server.
        keypair = self.cloud.get_keypair(server.key_name)
        yield from self.keypair(keypair)

        # The security groups and other network settings for the
        # server.
        for sg in server.security_groups:
            sg_data = self.cloud.get_security_group(sg.name)
            yield from self.security_group(sg_data)
        for net_name in server.networks:
            net_data = self.cloud.get_network(net_name)
            yield from self.network(net_data)
        # FIXME(dhellmann): Need to handle public IPs. Use auto_ip?

        # Any volumes attached to the server.
        # FIXME(dhellmann): The server needs to be stopped and the
        # volumes need to be detached before this step.
        vol_names = []
        # for vol in server.volumes:
        #     vol_data = self.cloud.get_volume(vol.id)
        #     vol_names.append(vol_data.name)
        #     yield from self.volume(vol_data, save_state)

        # The main portion of the server image.
        # FIXME(dhellmann): It looks like ceph-backed servers have an
        # image ID set to their volume or something? It's not a public
        # image visible through the glance API.
        if save_state:
            image_name = 'downpour-server-capture-{}'.format(server.id)
            image = self.cloud.get_image(image_name)
            if image:
                LOG.info('found existing capture of server %s in %s',
                         server.name, image_name)
            else:
                LOG.info('capturing server %s to image %s',
                         server.name, image_name)
                image = self.cloud.create_image_snapshot(
                    name=image_name,
                    server=server,
                    wait=True,
                )
        else:
            image = self.cloud.get_image(server.image.id)
        yield from self.image(image)

        server_data = {
            'name': server.name,
            'state': 'present',
            # Attach to the networks by name.
            'nics': list(server.networks.keys()),
            'image': image.name,
        }
        key_name = key_name or server.key_name
        if key_name:
            server_data['key_name'] = key_name
        if vol_names:
            server_data['volumes'] = vol_names
        yield {
            'name': 'Creating server {}'.format(server.name),
            'os_server': server_data,
            'register': self._mk_var_name('server'),
        }
        yield self._map_uuids('server', server.name, server.id, 'server.id')

    def image(self, image):
        filename = self._downloader.add_image(image)
        image_data = {
            'name': image.name,
            'container_format': image.container_format,
            'disk_format': image.disk_format,
            'min_disk': image.min_disk,
            'min_ram': image.min_ram,
        }
        # FIXME(dhellmann): handle ramdisk property?
        if filename:
            image_data['filename'] = filename
        yield {
            'name': 'Creating image {}'.format(image.name),
            'os_image': image_data,
            'register': self._mk_var_name('img'),
        }
        yield self._map_uuids('image', image.name, image.id, 'image.id')



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import logging
import sys

import os_client_config
import pkg_resources
import shade

from downpour import export
from downpour import query


def main():
    parser = argparse.ArgumentParser()
    config = os_client_config.OpenStackConfig()

    version_info = pkg_resources.get_distribution('os-downpour').version
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + version_info,
        help='show the program version and exit',
    )

    verbose_group = parser.add_mutually_exclusive_group()
    verbose_group.add_argument(
        '--verbose', '-v',
        action='count',
        dest='verbose_level',
        default=1,
        help='Increase verbosity of output. Can be repeated.',
    )
    verbose_group.add_argument(
        '-q', '--quiet',
        action='store_const',
        dest='verbose_level',
        const=0,
        help='Suppress output except warnings and errors.',
    )
    progress_group = parser.add_mutually_exclusive_group()
    progress_group.add_argument(
        '--progress', '-p',
        default=True,
        action='store_true',
        help='show download progress',
    )
    progress_group.add_argument(
        '--no-progress',
        dest='progress',
        action='store_false',
        help='do not show download progress',
    )

    config.register_argparse_arguments(parser, sys.argv, None)
    subparsers = parser.add_subparsers(title='commands')

    export.register_command(subparsers)
    query.register_command(subparsers)

    args = parser.parse_args(sys.argv[1:])

    cloud_config = config.get_one_cloud(options=(args, []))
    cloud = shade.OpenStackCloud(cloud_config=cloud_config)

    root_logger = logging.getLogger('')
    root_logger.setLevel(logging.DEBUG)
    console = logging.StreamHandler(sys.stderr)
    console_level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(args.verbose_level, logging.DEBUG)
    console.setLevel(console_level)
    formatter = logging.Formatter('[%(asctime)s] %(message)s')
    console.setFormatter(formatter)
    root_logger.addHandler(console)

    return args.func(cloud, config, args)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os.path

import yaml

from downpour import download
from downpour import resolver
from downpour import resources

LOG = logging.getLogger(__name__)


def register_command(subparsers):
    do_export = subparsers.add_parser(
        'export',
        help='export data',
    )
    do_export.add_argument(
        'resource_file',
        help='the name of the file listing resources to be exported',
    )
    do_export.add_argument(
        'output_path',
        default='.',
        nargs='?',
        help='the name of a directory to use for output file(s)',
    )
    do_export.set_defaults(func=export_data)


def export_data(cloud, config, args):
    output_path = args.output_path

    downloader = download.Downloader(
        output_path,
        cloud,
        use_progress_bar=(args.progress and (args.verbose_level >= 1)),
    )
    res = resolver.Resolver(cloud, downloader, output_path)
    tasks = []

    # Any tasks needed to initialize the output.
    tasks.extend(res.init_tasks())

    # Export independent resources. The resolver handles dependencies
    # automatically.
    to_export = resources.load(args.resource_file)

    for keypair_info in to_export.keypairs:
        keypair = cloud.get_keypair(keypair_info.name)
        tasks.extend(res.keypair(keypair))

    for image_info in to_export.images:
        image = cloud.get_image(image_info.name)
        tasks.extend(res.image(image))

    for volume_info in to_export.volumes:
        volume = cloud.get_volume(volume_info.name)
        tasks.extend(res.volume(volume, save_state=volume_info.save_state))

    for server_info in to_export.servers:
        server = cloud.get_server(server_info.name)
        tasks.extend(res.server(server,
                                save_state=server_info.save_state,
                                key_name=server_info.get('key_name')))

    playbook = [
        # The default playbook is configured to run instructions
        # locally to talk to the cloud API.
        {'hosts': 'localhost',
         'connection': 'local',
         'tasks': tasks,
         },
    ]
    playbook_filename = os.path.join(output_path, 'playbook.yml')
    with open(playbook_filename, 'w', encoding='utf-8') as fd:
        yaml.dump(playbook, fd, default_flow_style=False, explicit_start=True)
    LOG.info('wrote playbook to %s', playbook_filename)

    downloader.start()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from downpour import resources

LOG = logging.getLogger(__name__)


class ResourceFileEditor:

    def __init__(self, filename, save_state):
        self._filename = filename
        self._save_state = save_state
        self._resources = resources.load(filename, missing_ok=True)
        self._servers = set(s.name for s in self._resources.servers)
        self._images = set(i.name for i in self._resources.images)
        self._volumes = set(v.name for v in self._resources.volumes)

    def add_server(self, info):
        if info.name in self._servers:
            return
        LOG.info('found server %s to export', info.name)
        self._resources.servers.append({
            'name': info.name,
            'save_state': self._save_state,
            'key_name': info.key_name,
        })

    def add_image(self, info):
        if info.name in self._images:
            return
        LOG.info('found image %s to export', info.name)
        self._resources.images.append({
            'name': info.name,
        })

    def add_volume(self, info):
        if info.name in self._volumes:
            return
        LOG.info('found volume %s to export', info.name)
        self._resources.volumes.append({
            'name': info.name,
            'save_state': self._save_state,
        })

    def save(self):
        resources.save(self._filename, self._resources)


def register_command(subparsers):
    do_query = subparsers.add_parser(
        'query',
        help='query to build an export list',
    )
    do_query.add_argument(
        'resource_file',
        help='the name of the file listing resources to be updated',
    )
    do_query.add_argument(
        '--save-state',
        action='store_true',
        default=False,
        help='should the state of the server or volume be saved',
    )
    do_query.add_argument(
        '--server-name',
        action='append',
        default=[],
        help='pattern to match against server names',
    )
    do_query.add_argument(
        '--server-flavor',
        action='append',
        default=[],
        help='server flavor(s) to capture',
    )
    do_query.add_argument(
        '--image-name',
        action='append',
        default=[],
        help='pattern to match against image names',
    )
    do_query.add_argument(
        '--volume-name',
        action='append',
        default=[],
        help='pattern to match against volume names',
    )
    do_query.set_defaults(func=query_data)


def query_data(cloud, config, args):
    editor = ResourceFileEditor(args.resource_file, save_state=args.save_state)

    for pattern in args.server_name:
        LOG.info('searching for servers matching pattern %r', pattern)
        for info in cloud.search_servers(name_or_id=pattern):
            editor.add_server(info)

    for flavor in args.server_flavor:
        LOG.info('searching for servers using flavor %r', flavor)
        for info in cloud.search_servers(
                filters={'flavor': {'id': flavor}}):
            editor.add_server(info)

    for pattern in args.image_name:
        LOG.info('searching for images matching pattern %r', pattern)
        for info in cloud.search_images(name_or_id=pattern):
            editor.add_image(info)

    for pattern in args.volume_name:
        LOG.info('searching for available volumes matching pattern %r',
                 pattern)
        for info in cloud.search_volumes(
                name_or_id=pattern,
                filters={
                    'status': 'available',
                }):
            editor.add_volume(info)

    editor.save()



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os.path

import progressbar

LOG = logging.getLogger(__name__)


class ProgressBarDownloader:

    def __init__(self, output_path, max_value):
        self.output_path = output_path
        self.max_value = max_value
        self.bar = None
        self.fd = None
        self.amt_read = 0

    def __enter__(self):
        self.bar = progressbar.ProgressBar(
            widgets=[
                progressbar.Percentage(),
                ' ',
                progressbar.Bar(),
                progressbar.FileTransferSpeed(),
                ' ',
                progressbar.ETA(),
            ],
            max_value=self.max_value,
        )
        self.fd = open(self.output_path, 'wb')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fd.close()
        self.bar.finish()
        return False

    def write(self, block):
        if block:
            self.fd.write(block)
            self.amt_read += len(block)
            self.bar.update(self.amt_read)


class Downloader:

    def __init__(self, output_dir, cloud, use_progress_bar):
        self.output_dir = output_dir
        self.cloud = cloud
        self.use_progress_bar = use_progress_bar
        self._tasks = []

    def _add(self, resource_type, resource, output_path):
        LOG.info('scheduling download of %s %s', resource_type, resource.name)
        self._tasks.append((resource_type, resource, output_path))

    def add_image(self, image):
        base = image.name + '.dat'
        output_path = os.path.join(self.output_dir, base)
        self._add('image', image, output_path)
        return base

    def add_volume(self, volume):
        LOG.error('DO NOT KNOW HOW TO SAVE VOLUME STATE YET %s', volume.name)

    def start(self):
        # FIXME(dhellmann): start downloads in a separate thread or process
        for resource_type, resource, output_path in self._tasks:
            if os.path.exists(output_path):
                LOG.info(
                    'output file %s already exists, skipping download',
                    output_path,
                )
                continue
            if resource_type == 'image':
                LOG.info(
                    'downloading image %s to %s',
                    resource.name,
                    output_path,
                )
                if self.use_progress_bar:
                    with ProgressBarDownloader(output_path, resource.size) as out:
                        self.cloud.download_image(resource.name, output_file=out)
                else:
                    with open(output_path, 'wb') as out:
                        self.cloud.download_image(resource.name, output_file=out)
                LOG.info(
                    'downloaded image %s to %s',
                    resource.name,
                    output_path,
                )



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import munch

LOG = logging.getLogger(__name__)


def load(filename, missing_ok=False):
    "Read the file and return the parsed data in a consistent format."
    LOG.info('loading resource list from %s', filename)

    # Ensure the return value has a basic set of keys representing the
    # types of resources we expect to find.
    to_return = munch.Munch(
        servers=[],
        volumes=[],
        images=[],
        keypairs=[],
    )

    try:
        with open(filename, 'r', encoding='utf-8') as fd:
            contents = munch.Munch.fromYAML(fd.read())
    except FileNotFoundError:
        if not missing_ok:
            raise
    else:
        to_return.update(contents)

    # Ensure all entries have consistent sets of keys so the rest of
    # the app doesn't have to check every time it wants to use a
    # value.
    for s in to_return.servers:
        if 'save_state' not in s:
            s['save_state'] = True
    for s in to_return.volumes:
        if 'save_state' not in s:
            s['save_state'] = True

    return to_return


def save(filename, to_export):
    "Write the resources file."
    with open(filename, 'w', encoding='utf-8') as fd:
        fd.write(to_export.toYAML())
